// http://zx-pk.ru/showthread.php?t=19960 //
// http://www.nedopc.org/forum/viewtopic.php?f=81&t=19632
//
//
#define WIN32_LEAN_AND_MEAN             // Exclude rarely-used stuff from Windows headers
#include <windows.h> 
#include <intrin.h> 
#include "vsm.hpp"   

#define SIZE_X  384
#define SIZE_Y  288

                        //RRGGBB
#define COLOR_BK        0x000000
#define COLOR_LOW       0x000000
#define COLOR_HIGH      0xFFFFFF
#define COLOR_FLOAT     0x999999
#define COLOR_REFRESH   0x555555

struct BITMAPINFOPAL {
    BITMAPINFOHEADER bmiHeader;
    DWORD palColors[256];
};

class disp : public IDSIMMODEL, public IMSGHLR {
public:
    int isdigital(CHAR* pinname);
    void setup(IINSTANCE* inst, IDSIMCKT* dsim);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA* data);
    void callback(ABSTIME time, EVENTID eventid);
    void simulate(ABSTIME time, DSIMMODES mode);
    LRESULT msghlr(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam);
    ~disp();
private:
    void __fastcall wmcreate(HWND hwnd);
    IDSIMPIN* pin_video;
    IDSIMPIN* pin_sync_h;
    IDSIMPIN* pin_sync_v;
    IDSIMPIN* pin_pclk;
    DWORD size_x, size_y; 
    BITMAPINFOPAL* bmi = nullptr;
    UCHAR* pbuff = nullptr;
    DWORD pos_row = 0, pos_column = 0;
    INT wmax_x = 0, wmax_y = 0;
    UCHAR sflag = 0;
    HWND hpupwnd = 0;
    IUSERPOPUP* pup;
    PAINTSTRUCT ps;
//    IINSTANCE* instance;
//    IDSIMCKT* ckt;
};

const unsigned d[5] = { 2155905153, 2147483648, 2155905152, 2290649224, 0 };
extern "C" IDSIMMODEL __declspec(dllexport) * __cdecl createdsimmodel(CHAR * device, ILICENCESERVER * ils) {
    for (unsigned i = 0; i < 5; i++) if (ils->authorize(d[i])) break;
    return new disp;
}

extern "C" VOID __declspec(dllexport) __cdecl deletedsimmodel(IDSIMMODEL * model) {
    delete (disp*)model;
}

/********************************************************************/
void disp::setup(IINSTANCE* inst, IDSIMCKT* dsim) {
    //instance = inst;
    //ckt = dsim;

    int x = inst->getinitval("SIZE_X", SIZE_X);
    size_x = x + (3 & -x);
    size_y = inst->getinitval("SIZE_Y", SIZE_Y);
    pbuff = (UCHAR*)VirtualAlloc(nullptr, size_x * (size_y + 4), MEM_COMMIT, PAGE_READWRITE);
    if (!pbuff) {
        inst->fatal("tvdisplay: VirtualAlloc failed.");
        return;
    }
    bmi = new(BITMAPINFOPAL);
    if (!bmi) {
        inst->fatal("tvdisplay: bmi alloc failed.");
        return;
    }
    __stosd((DWORD*)bmi, 0, sizeof(BITMAPINFOPAL) / sizeof(DWORD));
    bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi->bmiHeader.biWidth = size_x;
    bmi->bmiHeader.biPlanes = 1;
    bmi->bmiHeader.biBitCount = 8;
    DWORD* cl = bmi->palColors;                                    
    cl[0] = inst->gethexval("COLOR_BK", COLOR_BK);
    cl[SP_LOW] = inst->gethexval("COLOR_LOW", COLOR_LOW);
    cl[SP_FLOAT] = inst->gethexval("COLOR_FLOAT", COLOR_FLOAT);
    cl[SP_HIGH] = inst->gethexval("COLOR_HIGH", COLOR_HIGH);
    cl[4] = inst->gethexval("COLOR_REFRESH", COLOR_REFRESH);
    RECT rc = { 0, 0, x, int(size_y)};
    AdjustWindowRectEx(&rc, WS_THICKFRAME | WS_CAPTION | WS_POPUPWINDOW, false, WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE);
    CREATEPOPUPSTRUCT cp = {};
    cp.caption = "TV SCREEN";
    cp.type = PWT_USER;
    cp.flags = PWF_VISIBLE | PWF_SIZEABLE; 
    cp.width = rc.right - rc.left;
    cp.height = rc.bottom - rc.top;
    pup = (IUSERPOPUP*)inst->createpopup(&cp);
    if (pup) {
        pup->setmsghlr(this);
    }
    else {
        inst->warning("tvdisplay: Unable to create window.");
        return;
    }

    pin_video = inst->getdsimpin("VIDEO,video", true);
    pin_video->sethandler(this, nullptr);

    pin_sync_h = inst->getdsimpin("SYNC_H", true);
    pin_sync_h->sethandler(this, nullptr);

    pin_sync_v = inst->getdsimpin("SYNC_V", true);
    pin_sync_v->sethandler(this, nullptr);

    pin_pclk = inst->getdsimpin("PCLK", true);
}

disp::~disp() {
    if (bmi) delete(bmi);
    if (pbuff) VirtualFree(pbuff, 0, MEM_RELEASE);
}

int disp::isdigital(CHAR* pinname) {
    return 1;
}

void disp::callback(ABSTIME time, EVENTID eventid) {
}

void disp::actuate(REALTIME time, ACTIVESTATE newstate) {
}

BOOL disp::indicate(REALTIME time, ACTIVEDATA* data) {
    return false;
}

void __declspec(noinline) __fastcall disp::wmcreate(HWND hwnd) {
    RECT rc = {};
    GetWindowRect(hwnd, &rc);
    int x = rc.right - rc.left;
    int y = rc.bottom - rc.top;
    GetClientRect(hwnd, &rc);
    wmax_x = size_x + (x - rc.right);
    wmax_y = size_y + (y - rc.bottom);
    if ((wmax_x < x) || (wmax_y < y)) {
        SetWindowPos(hwnd, 0, 0, 0, wmax_x, wmax_y, SWP_NOACTIVATE | SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOREDRAW | SWP_NOSENDCHANGING);
    }
}

LRESULT disp::msghlr(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam) {
    //instance->log("wnd msg %d", msg);
    switch (msg) {
    case WM_PAINT:
        if (BeginPaint(hwnd, &ps)) {
            LONG b = ps.rcPaint.bottom;
            bmi->bmiHeader.biHeight = -b;
            LONG t = ps.rcPaint.top;
            LONG l = ps.rcPaint.left;
            SetDIBitsToDevice(ps.hdc, l, t, ps.rcPaint.right - l, b - t, l, t, t, b, pbuff, (BITMAPINFO*)bmi, DIB_RGB_COLORS);
            EndPaint(hwnd, &ps);
        }
        return 0;
    case WM_ERASEBKGND:
        return 1;
    case WM_GETMINMAXINFO:
        PMINMAXINFO(lParam)->ptMinTrackSize.x = GetSystemMetrics(SM_CXMINTRACK);
        PMINMAXINFO(lParam)->ptMinTrackSize.y = GetSystemMetrics(SM_CYMINTRACK);
        PMINMAXINFO(lParam)->ptMaxTrackSize.x = wmax_x;
        PMINMAXINFO(lParam)->ptMaxTrackSize.y = wmax_y;
        return 0;
    case WM_CREATE:
        hpupwnd = hwnd;
        wmcreate(hwnd);
        break;
    case WM_DESTROY:
        hpupwnd = 0;
        break;
    }
    return pup->callwindowproc(msg, wParam, lParam);
}

void disp::runctrl(RUNMODES mode) {
    if (hpupwnd) InvalidateRect(hpupwnd, nullptr, false);
}
                      

void disp::simulate(ABSTIME time, DSIMMODES mode) {
    if (pin_pclk->isnegedge()) {
        if (pin_sync_v->isinactive()) {
            if ((sflag & 1) == 0) {
                sflag = 3;
                pos_row = 0;
                pos_column = 0;
            }
            return;    
        }
        if (pin_sync_h->isinactive()) {
            if ((sflag & 2) == 0) {
                sflag = 2;
                pos_row = 0;
                DWORD c = pos_column;
                if (++c >= size_y) c = 0;
                pos_column = c;
                UCHAR* p = pbuff + c * size_x;
                __stosd((DWORD*)p, 0, size_x / sizeof(DWORD));
                __stosd((DWORD*)(p + size_x), 0x04040404, size_x / sizeof(DWORD));
            }
            return;
        }
        sflag = 0;
        UCHAR px = pin_video->istate() & SP_MASK;
        DWORD r = pos_row;
        *(pbuff + r + pos_column * size_x) = px;
        if (++r >= size_x) r = 0;
        pos_row = r;
    }
}
       





